import inspect
import logging

logger = logging.getLogger(__name__)


class Extension:
    """
    An element to be bind into the Service Runner.
    It also binds sub-extensions (Extension inside another)
    """
    _runner = None
    _params = None

    def __new__(cls, *args, **kwargs):
        # Hack from Nameko's Extension to enable us to instantiate a new
        # Extension on Bind independent of the arguments it has on __init__
        inst = super(Extension, cls).__new__(cls)
        inst._params = (args, kwargs)
        return inst

    def bind(self, runner):
        """
        Binds the extension to the service runner and parent.
        :param Runner runner: Service Runner
        """
        if self._runner is not None:
            return self._runner
            # TODO Letting all extensions being shareable will impact in
            #  something?
            #raise AssertionError("This extension is already bound.")
        self._runner = runner
        # Binding sub-extensions
        for attr_name, extension in inspect.getmembers(
                self, lambda x: isinstance(x, Extension)):
            extension.bind(runner)

    def setup(self):
        """
        Called After the binding and before the system start
        """
        pass

    def start(self):
        """
        Commands the extension to start running.
        """
        pass

    def stop(self):
        """
        Commands the extension to do a graceful stop
        """
        pass

    def _clone(self):
        cls = type(self)
        return cls(*self._params[0], **self._params[1])

    @property
    def runner(self):
        if self._runner is None:
            raise RuntimeError("This extension is not binded. You must bind "
                               "it to a runner before trying to attempt this.")
        return self._runner



from concurrent.futures.process import _ExceptionWithTraceback
from importlib import import_module


class Worker:
    def __init__(self, function_path, dependencies, translators,
                 config, *fn_args, _meta=None, **fn_kwargs):
        self.function_path = function_path
        self.config = config
        self.fn_args = fn_args
        self.fn_kwargs = fn_kwargs
        self._dependencies = dependencies or {}
        self._translators = translators or []
        self.result = None
        self.exception = None
        self._meta = _meta or {} # Content shared by extensions

    def get_callable(self):
        *module, function = self.function_path.split('.')
        function = getattr(import_module('.'.join(module)), function)
        return function

    def call_dependencies(self, action, *args, **kwargs):
        ret = {}
        for name, dependency in self._dependencies.items():
            ret[name] = getattr(dependency, action)(*args, **kwargs)
        return ret

    def run(self):
        self.function = self.get_callable()
        self.call_dependencies('setup')
        dependencies = self.call_dependencies(
            'get_dependency', self
        )
        self.translated_args = self.fn_args
        self.translated_kwargs = self.fn_kwargs
        for translator in self._translators:
            self.translated_args, self.translated_kwargs = translator.translate(
                *self.fn_args, **self.fn_kwargs
            )
        self.call_dependencies('before_call', self)
        try:
            # Updating kwargs
            fn_kwargs = {**self.translated_kwargs, **dependencies}
            self.result = self.function(*self.translated_args, **fn_kwargs)
        except Exception as exc:
            if self.config.get('WORKER_MODE', 'thread') == 'process':
                # Trick from concurrent.futures to keep the traceback in
                # process type worker.
                self.exception = _ExceptionWithTraceback(exc, exc.__traceback__)
            else:
                self.exception = exc
            self.call_dependencies('after_call', self, self.result, exc)
        self.call_dependencies('after_call', self, self.result, None)
        return self


class CallbackWorker(Worker):
    def __init__(self, callback_function, original_worker):
        self.callback_function = callback_function
        self.original_worker = original_worker
        args = (*original_worker.translated_args, original_worker.exception)
        kwargs = original_worker.translated_kwargs
        # We use the already translated content but the dependencies will
        # have to be re-created due to the pickling problem of Processing.
        super(CallbackWorker, self).__init__(
            callback_function, original_worker._dependencies, None,
            original_worker.config, *args, _meta=original_worker._meta,
            **kwargs)



from micro_framework.extensions import Extension


class Dependency(Extension):
    """
    Kind of extension that injects a dependency into the function.
    """

    def bind(self, runner):
        self.config = runner.config

    def before_call(self, worker):
        """
        Treatment before the actual function call.
        """
        pass

    def get_dependency(self, worker):
        """
        Injects the dependency.
        """
        pass

    def after_call(self, worker, result, exc):
        """
        Cleanup after the function has finished or raised.
        """
        pass



import logging
import time
from datetime import datetime

from kombu import Exchange, Queue

from micro_framework.amqp.dependencies import dispatch
from micro_framework.amqp.entrypoints import EventListener, BaseEventListener
from micro_framework.dependencies import Dependency
from micro_framework.entrypoints import Entrypoint
from micro_framework.exceptions import BackOffLimitReached, \
    BackOffIntervalNotReached
from micro_framework.extensions import Extension


logger = logging.getLogger(__name__)


class BackOffContext:
    """
    Object to be passed as argument to functions that implement a BackOff.
    """
    def __init__(self, retry_count, last_retry):
        self.retry_count = retry_count
        self.last_retry = last_retry


class BackOffData(Dependency):
    def get_dependency(self, worker):
        backoff_meta = worker._meta.get('backoff', {})
        return BackOffContext(
            retry_count=backoff_meta.get('retry_count', 0),
            last_retry=backoff_meta.get('last_retry', 0)
        )


class BackOff(Extension):
    """
    Adds BackOff Capability to a failed route. It is responsible to firing
    retries when needed.
    """
    def __init__(self, max_retries=3, interval=60000, exception_list=None):
        """
        :param int max_retries: Total number of retry attempts.
        :param interval: Interval between retries in micro seconds
        :param exception_list: list of exceptions that enable retries.
        """
        self.max_retries = max_retries
        self.interval = interval # Microseconds
        self.exception_list = exception_list
        self.route = None

    def get_interval(self):
        return self.interval

    def get_max_retries(self):
        return self.max_retries

    def get_exception_list(self):
        return self.exception_list

    def exception_in_list(self, exception):
        if not self.exception_list:
            return True
        return type(exception) in self.exception_list

    def retry_count_is_lower(self, context):
        retry_count = 0
        if context:
            retry_count = context['retry_count']
        if retry_count < self.get_max_retries():
            return True
        return False

    def interval_passed(self, context):
        last_retry = 0
        if context:
            last_retry = context['last_retry']
        if last_retry + self.interval / 1000 < datetime.utcnow().timestamp():
            return True
        return False

    def get_context(self, worker):
        return worker._meta.get('backoff', None)

    def can_retry(self, worker):
        """
        Check if the given worker is eligible to be sent to the retry queue.
        :param Worker worker: The resulting worker with exception
        :return bool:
        """
        backoff_context = self.get_context(worker)
        return all(
            [self.retry_count_is_lower(backoff_context),
             self.exception_in_list(worker.exception),
             self.interval_passed(backoff_context)
             ]
        )

    def update_meta(self, worker):
        # The arguments should be serializable....
        # We need to pop the context since it doesn't belong to the original
        # fn_kwargs
        context = self.get_context(worker)
        retry_count = 1
        if context is not None:
            retry_count = context['retry_count'] + 1
        meta = worker._meta or {}
        updated_backoff = {
            'retry_count': retry_count,
            'last_retry': datetime.utcnow().timestamp()
        }
        meta.setdefault('backoff', {}).update(updated_backoff)
        return meta

    def execute_retry(self, updated_worker):
        """
        Hook method to implement the retry logic. Ex: AsyncBackOff bellow.

        :param Worker updated_worker: A Worker with updated BackOff _meta
        """
        raise NotImplementedError()

    def retry(self, worker):
        """
        Called to invoke the retry action.
        It updates the backoff _meta attribute of the worker and send it
        :param Worker worker: the result worker with information to be used
        in the retry such as the fn_args and kwargs.
        """
        updated_meta = self.update_meta(worker)
        worker._meta = updated_meta
        logger.info(f"{worker.function_path} failed, sending retry "
                    f"{updated_meta['backoff']['retry_count']} of "
                    f"{self.max_retries}")
        self.execute_retry(worker)

    def bind_route(self, route):
        self.route = route


class AsyncBackOff(BackOff, BaseEventListener):
    """
    BackOff that implements an Event Listener entrypoint which is responsible
    to manage the backoff of a specific route.

    Since it is an entrypoint, the binded route needs to be a copy, to enable
    us binding this entrypoint to it, instead of the original entrypoint.


    # TODO Maybe adding multiple entrypoints to a route would make this implementation cleaner
    """
    def interval_passed(self, context):
        return True # RabbitMQ handles this.

    def mount_backoff_payload(self, worker):
        return {
            'fn_args': worker.fn_args,
            'fn_kwargs': worker.fn_kwargs,
            '_meta': worker._meta
        }

    def execute_retry(self, updated_worker):
        # We fire the event to a queue known by the instantiated entrypoint
        retry_payload = self.mount_backoff_payload(updated_worker)
        dispatch(
            self.runner.config['AMQP_URI'],
            self.service_name,
            self.get_event_name(),
            retry_payload,
            exchange=self.get_exchange(),
            headers={'x-delay': self.get_interval()}
        )

    def bind_route(self, route):
        self.route = route._clone()
        self.route._entrypoint = self
        self.route.bind(self.runner)
        self.route.entrypoint.bind_to_route(self.route)

    def get_event_name(self):
        return f"{self.route.function_path}_backoff"

    def get_exchange(self):
        return Exchange(
            name=self.service_name+'_backoff', type='x-delayed-message',
            arguments={'x-delayed-type': 'direct'}
        )

    def get_queue(self):
        return Queue(
            name=f"{self.get_event_name()}_queue",
            exchange=self.get_exchange(), routing_key=self.get_event_name()
        )

    def new_entry(self, message, payload):
        # Converting the BackOff payload to the original payload and adding a
        # BackOff Context to it as kwargs
        fn_args = payload['fn_args']
        fn_kwargs = payload['fn_kwargs']
        _meta = payload['_meta']
        self.call_route(message, *fn_args, _meta=_meta, **fn_kwargs)

    @property
    def service_name(self):
        return self.runner.config['SERVICE_NAME']



import logging
import time
import uuid
from datetime import datetime
from timeit import default_timer as timer

from micro_framework.exceptions import PoolStopped, ExtensionIsStopped
from micro_framework.extensions import Extension

logger = logging.getLogger(__name__)


class Entrypoint(Extension):
    route = None

    def bind_to_route(self, route):
        self.route = route

    def call_route(self, entry_id, *args, _meta=None, **kwargs):
        try:
            self.route.start_route(entry_id, *args, _meta=_meta, **kwargs)
        except PoolStopped:  # Unexpected error and we want to
            logger.info("New entry called when pool was already "
                        "signalled to stop.")
            self.on_failure(entry_id)
        except Exception:
            logger.exception("Failure when trying to start route for "
                             f"entrypoint: {self}")

    def on_finished_route(self, entry_id, worker):
        pass

    def on_failure(self, entry_id):
        pass


class TimeEntrypoint(Entrypoint):
    def __init__(self, interval):
        self.interval = interval
        self.running = False

    def start(self):
        self.running = True
        self.runner.spawn_extension(self, self.run)

    def run(self):
        while self.running:
            t1 = timer()
            entry_id = uuid.uuid4()
            now = datetime.utcnow()
            try:
                self.call_route(entry_id, now.isoformat())
            except (ExtensionIsStopped, PoolStopped):
                pass
            elapsed_time = timer() - t1
            sleep_time = self.interval - elapsed_time
            if sleep_time > 0:
                time.sleep(sleep_time)

    def stop(self):
        self.running = False



class FrameworkException(Exception):
    pass


class PoolStopped(Exception):
    pass

class ExtensionIsStopped(FrameworkException):
    pass


class BackOffLimitReached(FrameworkException):
    pass


class BackOffIntervalNotReached(FrameworkException):
    pass




# TODO get configurations somehow
import inspect
import logging.config
import time

from micro_framework.entrypoints import Entrypoint
from micro_framework.extensions import Extension
from micro_framework.spawners.multiprocess import ProcessSpawner, \
    ProcessPoolSpawner
from micro_framework.spawners.thread import ThreadSpawner, ThreadPoolSpawner


# Greedy Spawner means that the runner will consume all received tasks so far
# when a stop signal is sent.

# The non-greedy spawner will finish the current tasks and skip all pending
# ones. (Entrypoints with acknowledge are good candidates for this)
SPAWNERS = {
    'thread': ThreadSpawner, # Will
    'process': ProcessSpawner,
    'greedy_thread': ThreadPoolSpawner, # Will consume all received tasks on stop
    'greedy_process': ProcessPoolSpawner, # Will consume all received tasks
}

logger = logging.getLogger(__name__)


class Runner:
    def __init__(self, routes, config):
        self.config = config
        self.routes = routes
        self.worker_mode = config.get('WORKER_MODE', 'thread')
        self.spawner = SPAWNERS[self.worker_mode](config['MAX_WORKERS'])
        self.extension_spawner = ThreadSpawner()
        self.is_running = False
        self.spawned_workers = {}
        self.spawned_threads = {}
        self.bind_extensions()
        self.entrypoints = set()
        self.extra_extensions = set()
        self._find_extensions()
        logger.debug(f"Extensions: {self.extensions}")

    def _call_extensions_action(self, action, extension_set=None):
        if extension_set is None:
            extension_set = self.extensions
        for extension in extension_set:
            logger.debug(f"Calling {action} for {extension}")
            getattr(extension, action)()

    def _find_extensions(self):
        for route in self.routes:
            extensions = find_extensions(route)
            for extension in extensions:
                self.register_extension(extension)

    @property
    def extensions(self):
        return {*self.routes, *self.entrypoints, *self.extra_extensions}

    def bind_extensions(self):
        for route in self.routes:
            if not isinstance(route, Extension):
                raise TypeError("Only Extensions should be added as a "
                                "entrypoints.")
            route.bind(self)

    def register_extension(self, extension):
        if isinstance(extension, Entrypoint):
            self.entrypoints.add(extension)
        else:
            self.extra_extensions.add(extension)

    def start(self):
        logger.info(
            f"Starting Runner with {self.config['MAX_WORKERS']}"
            f" {self.worker_mode} workers"
        )
        # First we tell the routes to bind to their extensions
        self._call_extensions_action(
            'bind_to_extensions', extension_set=self.routes
        )
        logger.debug(f"Extensions: {self.extensions}")
        # Once we have all extensions detected:
        # Setup Routes first then entrypoints last and extra extensions last.
        self._call_extensions_action('setup', extension_set=self.routes)
        self._call_extensions_action('setup', extension_set=self.entrypoints)
        self._call_extensions_action('setup', extension_set=self.extra_extensions)

        self._call_extensions_action('start')
        self.is_running = True
        while self.is_running:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                self.stop()
            except Exception:
                self.stop()
                raise
        logger.info("Runner stopped")

    def stop(self, *args):
        self.is_running = False
        logger.info("Stopping all extensions and workers")
        # Stopping Entrypoints First
        self._call_extensions_action('stop', extension_set=self.routes)
        # Stop Workers
        logger.debug("Stopping workers")
        self.spawner.stop(wait=True)
        logger.debug("Stopping extra extensions")
        self._call_extensions_action('stop',
                                     extension_set=self.extra_extensions)
        # Stop running extensions
        logger.debug("Stopping any still running extensions thread")
        self.extension_spawner.stop(wait=False)

    def spawn_worker(self, worker, *fn_args, callback=None, **fn_kwargs):
        """
        Spawn a new worker to execute a function from the calling route.

        :param Route route: The route that called it
        :param fn_args: Arguments of the function to be called.
        :return Future: A Future object of the task.
        """
        future = self.spawner.spawn(
            worker.run, callback, *fn_args, **fn_kwargs
        )
        self.spawned_workers[future] = worker
        return future

    def spawn_extension(self, extension, target_fn, *fn_args, callback=None,
                        **fn_kwargs):
        """
        Any extension that needs a separate thread should call this method to
        start a new thread for it.
        :param Extension extension: An extension subclass
        :param function target_fn: the function to be executed.
        :param function callback: A callback function to handle the
        result/error
        :param fn_args: Any argument to be passed to the target function.
        :param fn_kwargs: Any named argument to be passed to the target
        function.
        :return Future: A future object of the task.
        """
        if callback is None:
            callback = self.extension_thread_callback
        future = self.extension_spawner.spawn(
            target_fn, callback, *fn_args, **fn_kwargs
        )
        self.spawned_threads[future] = extension
        return future

    def extension_thread_callback(self, future):
        """
        Handle the result of an extension when it's worker is finished either
        by an error or simply shutting down.
        """
        if future.exception():
            print("Exception in thread.")
            self.stop()
            raise future.exception()


def find_extensions(cls):
    extensions = set()
    for name, extension in inspect.getmembers(
            cls, lambda x: isinstance(x, Extension)):
        extensions.add(extension)
        extensions.update(find_extensions(extension))
    return extensions



import setuptools

setuptools.setup(
    packages=['micro_framework'],
    name='micro-framework',
    url='https://github.com/Mendes11/micro_framework',
    version='1.0.2',
    description='Framework to create CPU or IO bound microservices.',
    long_description= 'file: README.rst',
    author = 'Rafael Mendes Pacini Bachiega',
    author_email = 'rafaelmpb11@hotmail.com',
    classifiers =
    ['Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8'],
    install_requires=[
        "kombu==4.6.8",
    ],
)



import logging

from functools import partial

from micro_framework.exceptions import ExtensionIsStopped
from micro_framework.extensions import Extension
from micro_framework.workers import Worker, CallbackWorker

logger = logging.getLogger(__name__)


class Route(Extension):
    """
    Base Extension Destined to link an endpoint to an entry function.

    Protocol-Specific routes should implement this class and extend its
    features.

    In order to keep the framework code decoupled from the to-be executed
    functions, we pass only the path of the function so it will be imported
    only when a new worker is spawned, making any setup from imports and etc
    (Django i'm looking to you) run in the spawned worker.
    """
    worker_class = Worker

    def __init__(self, function_path, entrypoint, dependencies=None,
                 translators=None, worker_class=None, backoff=None):
        self._entrypoint = entrypoint
        self._dependencies = dependencies or {}
        self._translators = translators or []
        self._function_path = function_path
        self.stopped = False
        self.current_workers = {}
        self.worker_class = worker_class or self.worker_class
        self._backoff = backoff

    def handle_finished_worker(self, entry_id, worker):

        if worker.exception and self.backoff and self.backoff.can_retry(worker):
            self.backoff.retry(worker)  # Retry and let the entrypoint finish
        self.entrypoint.on_finished_route(entry_id, worker)
        if worker.exception:
            raise worker.exception

    def worker_result(self, entry_id, future):
        logger.debug(f"{self} Received a worker result.")
        # Cleaning runner
        self.runner.spawned_workers.pop(future)
        if future.exception():
            # Unhandled exception propagate it to kill thread.
            raise future.exception()
        worker = future.result()
        self.handle_finished_worker(entry_id, worker)

    def get_worker_instance(self, *fn_args, _meta=None, **fn_kwargs):
        return self.worker_class(
            self.function_path, self.dependencies.copy(),
            self.translators.copy(), self.runner.config, *fn_args, _meta=_meta,
            **fn_kwargs
        )

    def run_worker(self, entry_id, worker, callback=None):
        if callback is None:
            callback = partial(self.worker_result, entry_id)
        future = self.runner.spawn_worker(worker)
        future.add_done_callback(callback)

    def start_route(self, entry_id, *fn_args, _meta=None, **fn_kwargs):
        if self.stopped:
            raise ExtensionIsStopped()
        worker = self.get_worker_instance(*fn_args, _meta=_meta, **fn_kwargs)
        self.run_worker(entry_id, worker)
        return worker

    def bind_to_extensions(self):
        self.entrypoint.bind_to_route(self)
        if self.dependencies:
            for name, dependency in self.dependencies.items():
                dependency.bind(self.runner) # TODO Maybe binding later to route only
        if self.backoff is not None:
            # Link this route to the given backoff class
            self.backoff.bind_route(self)

    def stop(self):
        self.stopped = True

    @property
    def dependencies(self):
        return self._dependencies

    @property
    def function_path(self):
        return self._function_path

    @property
    def translators(self):
        return self._translators

    @property
    def entrypoint(self):
        return self._entrypoint

    @property
    def backoff(self):
        return self._backoff

    def __str__(self):
        return f'{self.__class__.__name__} -> {self.function_path}'

    def __repr__(self):
        return self.__str__()


class CallbackRoute(Route):
    callback_worker_class = CallbackWorker

    def __init__(self, *args, callback_function_path,
                 callback_worker_class=None, **kwargs):
        super(CallbackRoute, self).__init__(*args, **kwargs)
        self.callback_function_path = callback_function_path
        self.callback_worker_class = callback_worker_class or self.callback_worker_class

    def handle_finished_callback_worker(self, entry_id, callback_worker):
        self.entrypoint.on_finished_route(
            entry_id, worker=callback_worker.original_worker
        )
        if callback_worker.exception:
            raise callback_worker.exception

    def callback_worker_result(self, entry_id, future):
        logger.debug(f"{self} received a callback worker result.")
        self.runner.spawned_workers.pop(future)
        if future.exception():
            raise future.exception()
        callback_worker = future.result()
        self.handle_finished_callback_worker(entry_id, callback_worker)

    def get_callback_worker_instance(self, original_worker):
        return self.callback_worker_class(
            self.callback_function_path, original_worker
        )

    def start_callback_route(self, entry_id, worker):
        logger.info(f"{self.function_path} failed. Starting callback route.")
        callback_worker = self.get_callback_worker_instance(worker)
        callback = partial(self.callback_worker_result, entry_id)
        self.run_worker(entry_id, callback_worker, callback=callback)
        return callback_worker

    def handle_finished_worker(self, entry_id, worker):
        # Callback route will only be called if retry is not valid.
        if worker.exception and self.backoff and self.backoff.can_retry(worker):
            self.backoff.retry(worker)  # Retry and let the entrypoint finish
            return self.entrypoint.on_finished_route(entry_id, worker)
        elif worker.exception and self.callback_function_path:
            logger.debug(
                "Finished worker has an exception, calling callback route."
            )
            self.start_callback_route(entry_id, worker)
            raise worker.exception
        else:
            super(CallbackRoute, self).handle_finished_worker(entry_id, worker)



import json


class MessageTranslator:
    """
    Simple class receives a message and returns a transformation of it.
    """
    def get_message(self, *args, **kwargs):
        return args[0]

    def set_message(self, message, *args, **kwargs):
        list(args)[0] = message
        return args, kwargs

    def translate_message(self, message):
        raise NotImplementedError()

    def translate(self, *args, **kwargs):
        message = self.get_message(*args, **kwargs)
        translated = self.translate(message)
        return self.set_message(translated, *args, **kwargs)


class JSONTranslator(MessageTranslator):
    def translate_message(self, message):
        if isinstance(message, dict):
            return message
        return json.loads(message)
